import { Directive, EventEmitter, Input, NgZone, Optional, Output, SkipSelf, } from '@angular/core';
import { BehaviorSubject } from 'rxjs';
import { filter } from 'rxjs/operators';
import { Color } from 'three';
import { AnimationStore, CanvasStore, EventsStore, InstancesStore, } from '../stores';
import { applyProps } from '../utils';
import { AnimationLoopParticipant } from './animation-loop-participant.abstract';
export class ThreeObject3d extends AnimationLoopParticipant {
    constructor(canvasStore, instancesStore, eventsStore, animationStore, ngZone, parentObjectDirective) {
        super(animationStore, ngZone);
        this.canvasStore = canvasStore;
        this.instancesStore = instancesStore;
        this.eventsStore = eventsStore;
        this.animationStore = animationStore;
        this.ngZone = ngZone;
        this.parentObjectDirective = parentObjectDirective;
        this.castShadow = false;
        this.receiveShadow = false;
        this.visible = true;
        this.matrixAutoUpdate = true;
        this.appendMode = 'immediate';
        // events
        this.click = new EventEmitter();
        this.contextmenu = new EventEmitter();
        this.dblclick = new EventEmitter();
        this.pointerup = new EventEmitter();
        this.pointerdown = new EventEmitter();
        this.pointerover = new EventEmitter();
        this.pointerout = new EventEmitter();
        this.pointerenter = new EventEmitter();
        this.pointerleave = new EventEmitter();
        this.pointermove = new EventEmitter();
        this.pointermissed = new EventEmitter();
        this.pointercancel = new EventEmitter();
        this.wheel = new EventEmitter();
        this.$object3d = new BehaviorSubject(null);
        this.ready = this.object3d$.pipe(filter(Boolean));
    }
    get object3d$() {
        return this.$object3d.asObservable();
    }
    ngOnChanges(changes) {
        if (this.object3d) {
            this.applyCustomProps(changes);
        }
    }
    init() {
        this.ngZone.runOutsideAngular(() => {
            this.initObject();
            if (this.object3d) {
                this.applyCustomProps();
                applyProps(this.object3d, {
                    __ngt: {
                        canvasStateGetter: () => this.canvasStore.getImperativeState(),
                        eventsStateGetter: () => this.eventsStore.getImperativeState(),
                        handlers: this.applyEvents(),
                    },
                });
                this.instancesStore.saveObject(this.object3d);
                this.appendToParent();
                this.objectReady();
            }
        });
    }
    addToScene() {
        const { scene } = this.canvasStore.getImperativeState();
        if (scene) {
            scene.add(this.object3d);
        }
    }
    addToParent() {
        if (this.parentObjectDirective) {
            this.parentObjectDirective.object3d.add(this.object3d);
        }
        else {
            this.addToScene();
        }
    }
    remove() {
        if (this.appendTo) {
            this.appendTo.remove(this.object3d);
        }
        else if (this.parentObjectDirective && this.appendMode === 'immediate') {
            this.parentObjectDirective.object3d.remove(this.object3d);
        }
        else {
            const { scene } = this.canvasStore.getImperativeState();
            if (scene) {
                scene.remove(this.object3d);
            }
        }
        this.object3d.clear();
    }
    objectReady() {
        this.$object3d.next(this.object3d);
        this.participate(this.object3d);
    }
    applyCustomProps(inputChanges) {
        this.ngZone.runOutsideAngular(() => {
            const customProps = {
                castShadow: this.castShadow,
                receiveShadow: this.receiveShadow,
                visible: this.visible,
                matrixAutoUpdate: this.matrixAutoUpdate,
            };
            if (this.name) {
                customProps['name'] = this.name;
            }
            if (this.position) {
                customProps['position'] = this.position;
            }
            if (this.rotation) {
                customProps['rotation'] = this.rotation;
            }
            else if (this.quaternion) {
                customProps['quaternion'] = this.quaternion;
            }
            if (this.scale) {
                customProps['scale'] = this.scale;
            }
            if (this.userData) {
                customProps['userData'] = this.userData;
            }
            if (this.color) {
                this.color = Array.isArray(this.color)
                    ? new Color(...this.color)
                    : new Color(this.color);
                if (!this.canvasStore.getImperativeState().isLinear) {
                    this.color.convertSRGBToLinear();
                }
                customProps['color'] = this.color;
            }
            if (inputChanges) {
                for (const [inputName, inputChange] of Object.entries(inputChanges)) {
                    if (!inputChange.isFirstChange() ||
                        [
                            'name',
                            'position',
                            'rotation',
                            'quaternion',
                            'scale',
                            'userData',
                            'color',
                        ].includes(inputName) // skip 7 common inputs
                    ) {
                        continue;
                    }
                    customProps[inputName] = inputChange.currentValue;
                }
            }
            applyProps(this.object3d, customProps);
            this.object3d.updateMatrix();
        });
    }
    applyEvents() {
        const handlers = {};
        [
            'click',
            'contextmenu',
            'dblclick',
            'pointerup',
            'pointerdown',
            'pointerover',
            'pointerout',
            'pointerenter',
            'pointerleave',
            'pointermove',
            'pointermissed',
            'pointercancel',
            'wheel',
        ].forEach((eventName) => {
            if (this[eventName].observers.length) {
                handlers[eventName] = (event) => {
                    this.ngZone.run(() => {
                        this[eventName].emit(event);
                    });
                };
            }
        });
        return handlers;
    }
    appendToParent() {
        if (this.appendTo) {
            this.appendTo.add(this.object3d);
            return;
        }
        if (this.appendMode === 'root') {
            this.addToScene();
            return;
        }
        if (this.appendMode === 'immediate') {
            this.addToParent();
        }
    }
    ngOnDestroy() {
        super.ngOnDestroy();
        this.ngZone.runOutsideAngular(() => {
            if (this.object3d) {
                this.remove();
                this.instancesStore.removeObject(this.object3d.uuid);
                this.animationStore.unregisterAnimationEffect(this.object3d.uuid);
            }
        });
    }
}
ThreeObject3d.decorators = [
    { type: Directive }
];
ThreeObject3d.ctorParameters = () => [
    { type: CanvasStore },
    { type: InstancesStore },
    { type: EventsStore },
    { type: AnimationStore },
    { type: NgZone },
    { type: ThreeObject3d, decorators: [{ type: Optional }, { type: SkipSelf }] }
];
ThreeObject3d.propDecorators = {
    name: [{ type: Input }],
    position: [{ type: Input }],
    rotation: [{ type: Input }],
    quaternion: [{ type: Input }],
    scale: [{ type: Input }],
    color: [{ type: Input }],
    userData: [{ type: Input }],
    castShadow: [{ type: Input }],
    receiveShadow: [{ type: Input }],
    visible: [{ type: Input }],
    matrixAutoUpdate: [{ type: Input }],
    appendMode: [{ type: Input }],
    appendTo: [{ type: Input }],
    click: [{ type: Output }],
    contextmenu: [{ type: Output }],
    dblclick: [{ type: Output }],
    pointerup: [{ type: Output }],
    pointerdown: [{ type: Output }],
    pointerover: [{ type: Output }],
    pointerout: [{ type: Output }],
    pointerenter: [{ type: Output }],
    pointerleave: [{ type: Output }],
    pointermove: [{ type: Output }],
    pointermissed: [{ type: Output }],
    pointercancel: [{ type: Output }],
    wheel: [{ type: Output }],
    ready: [{ type: Output }]
};
//# sourceMappingURL=object-3d.abstract.js.map