import { __rest } from 'tslib';
import { Injectable, EventEmitter, Directive, NgZone, Output, Optional, SkipSelf, Input } from '@angular/core';
import { isObservable, combineLatest, Subject, BehaviorSubject } from 'rxjs';
import { tap, withLatestFrom, skip, filter, mapTo } from 'rxjs/operators';
import { Vector3, Vector2, Clock, WebGLRenderer, PCFSoftShadowMap, ACESFilmicToneMapping, sRGBEncoding, Scene, Camera, OrthographicCamera, PerspectiveCamera, Raycaster, Object3D, Color, Material, BufferGeometry } from 'three';
import { ComponentStore } from '@ngrx/component-store';

/**
 * Generate a random Id or an Event's uuid
 *
 * @param {ThreeIntersection} event
 *
 * @internal
 * @private
 */
function makeId(event) {
    if (event) {
        return (event.eventObject || event.object).uuid + '/' + event.index;
    }
    return id();
}
const cache = {};
/**
 * Generates a short id.
 *
 * Description:
 *  A 5-character alphanumeric sequence (364 = 1.6 million)
 *  This should only be used for JavaScript specific models.
 *  http://stackoverflow.com/questions/6248666/how-to-generate-short-uid-like-ax4j9z-in-js
 *
 *  Example: `aebgf`
 */
function id() {
    let newId = ('0000' + ((Math.random() * Math.pow(36, 4)) << 0).toString(36)).slice(-4);
    newId = `a${newId}`;
    // ensure not already used
    if (!cache[newId]) {
        cache[newId] = true;
        return newId;
    }
    return id();
}

/**
 * From r3f https://github.com/pmndrs/react-three-fiber/blob/master/packages/fiber/src/core/events.ts
 */
function createEvents(canvasStateGetter, eventsStateGetter, instancesStateGetter) {
    const temp = new Vector3();
    /** Sets up defaultRaycaster */
    function prepareRay(event) {
        var _a, _b;
        const { raycaster, mouse, camera, clock, renderer, scene, internal: { size }, } = canvasStateGetter();
        const delta = clock.getDelta();
        if (raycaster && camera && renderer && scene) {
            const { offsetX, offsetY } = (_b = (_a = raycaster.computeOffsets) === null || _a === void 0 ? void 0 : _a.call(raycaster, event, {
                camera,
                clock,
                size,
                renderer,
                scene,
                delta,
            })) !== null && _b !== void 0 ? _b : event;
            const { width, height } = size;
            mouse.set((offsetX / width) * 2 - 1, -(offsetY / height) * 2 + 1);
            raycaster.setFromCamera(mouse, camera);
        }
    }
    /** Calculates delta */
    function calculateDistance(event) {
        const { internal: { initialClick }, } = eventsStateGetter();
        const dx = event.offsetX - initialClick[0];
        const dy = event.offsetY - initialClick[1];
        return Math.round(Math.sqrt(dx * dx + dy * dy));
    }
    /** Returns true if an instance has a valid pointer-event registered, this excludes scroll, clicks etc */
    function filterPointerEvents(objects) {
        return objects.filter((obj) => ['move', 'over', 'enter', 'out', 'leave'].some((name) => { var _a, _b; return (_b = (_a = obj.__ngt) === null || _a === void 0 ? void 0 : _a.handlers) === null || _b === void 0 ? void 0 : _b[('pointer' + name)]; }));
    }
    function intersect(filter) {
        var _a;
        const { raycaster } = canvasStateGetter();
        const { objects } = instancesStateGetter();
        // Skip event handling when noEvents is set
        if (!raycaster || (raycaster && !raycaster.enabled))
            return [];
        const seen = new Set();
        const intersections = [];
        // Allow callers to eliminate event objects
        const eventsObjects = filter
            ? filter(Object.values(objects))
            : Object.values(objects);
        // Intersect known handler objects and filter against duplicates
        const intersects = raycaster
            .intersectObjects(eventsObjects, true)
            .filter((item) => {
            const id = makeId(item);
            if (seen.has(id))
                return false;
            seen.add(id);
            return true;
        });
        // https://github.com/mrdoob/three.js/issues/16031
        // Allow custom userland intersect sort order
        // if (raycaster.filter) intersects = raycaster.filter(intersects, state);
        for (const intersect of intersects) {
            let eventObject = intersect.object;
            // Bubble event up
            while (eventObject) {
                const handlers = (_a = eventObject.__ngt) === null || _a === void 0 ? void 0 : _a.handlers;
                if (handlers)
                    intersections.push(Object.assign(Object.assign({}, intersect), { eventObject }));
                eventObject = eventObject.parent;
            }
        }
        return intersections;
    }
    /**  Creates filtered intersects and returns an array of positive hits */
    function patchIntersects(intersections, event) {
        const { internal: { captured }, } = eventsStateGetter();
        // If the interaction is captured take that into account, the captured event has to be part of the intersects
        if (captured && event.type !== 'click' && event.type !== 'wheel') {
            captured.forEach((captured) => {
                if (!intersections.find((hit) => hit.eventObject === captured.eventObject))
                    intersections.push(captured);
            });
        }
        return intersections;
    }
    /**  Handles intersections by forwarding them to handlers */
    function handleIntersects(intersections, event, callback) {
        const { raycaster, mouse, camera } = canvasStateGetter();
        const { internal: eventsInternal } = eventsStateGetter();
        // If anything has been found, forward it to the event listeners
        if (intersections.length && camera) {
            const unprojectedPoint = temp.set(mouse.x, mouse.y, 0).unproject(camera);
            const delta = event.type === 'click' ? calculateDistance(event) : 0;
            const releasePointerCapture = (id) => event.target.releasePointerCapture(id);
            const localState = { stopped: false, captured: false };
            for (const hit of intersections) {
                const setPointerCapture = (id) => {
                    // If the hit is going to be captured flag that we're in captured state
                    if (!localState.captured) {
                        localState.captured = true;
                        // The captured hit array is reset to collect hits
                        eventsInternal.captured = [];
                    }
                    // Push hits to the array
                    if (eventsInternal.captured)
                        eventsInternal.captured.push(hit);
                    // Call the original event now
                    event.target.setPointerCapture(id);
                };
                // Add native event props
                const extractEventProps = {};
                for (const prop in Object.getPrototypeOf(event)) {
                    // noinspection JSUnfilteredForInLoop
                    extractEventProps[prop] = event[prop];
                }
                const raycastEvent = Object.assign(Object.assign(Object.assign({}, hit), extractEventProps), { intersections, stopped: localState.stopped, delta,
                    unprojectedPoint, ray: raycaster ? raycaster.ray : null, camera: camera, 
                    // Hijack stopPropagation, which just sets a flag
                    stopPropagation: () => {
                        // https://github.com/pmndrs/react-three-fiber/issues/596
                        // Events are not allowed to stop propagation if the pointer has been captured
                        const cap = eventsInternal.captured;
                        if (!cap ||
                            cap.find((h) => h.eventObject.id === hit.eventObject.id)) {
                            raycastEvent.stopped = localState.stopped = true;
                            // Propagation is stopped, remove all other hover records
                            // An event handler is only allowed to flush other handlers if it is hovered itself
                            if (eventsInternal.hovered.size &&
                                Array.from(eventsInternal.hovered.values()).find((i) => i.eventObject === hit.eventObject)) {
                                // Objects cannot flush out higher up objects that have already caught the event
                                const higher = intersections.slice(0, intersections.indexOf(hit));
                                cancelPointer([...higher, hit]);
                            }
                        }
                    }, target: Object.assign(Object.assign({}, event.target), { setPointerCapture, releasePointerCapture }), currentTarget: Object.assign(Object.assign({}, event.currentTarget), { setPointerCapture,
                        releasePointerCapture }), sourceEvent: event });
                // Call subscribers
                callback(raycastEvent);
                // Event bubbling may be interrupted by stopPropagation
                if (localState.stopped)
                    break;
            }
        }
        return intersections;
    }
    function cancelPointer(hits) {
        const { internal } = eventsStateGetter();
        Array.from(internal.hovered.values()).forEach((hoveredObj) => {
            var _a, _b, _c;
            // When no objects were hit or the the hovered object wasn't found underneath the cursor
            // we call onPointerOut and delete the object from the hovered-elements map
            if (!hits.length ||
                !hits.find((hit) => hit.object === hoveredObj.object && hit.index === hoveredObj.index)) {
                const eventObject = hoveredObj.eventObject;
                const handlers = (_a = eventObject.__ngt) === null || _a === void 0 ? void 0 : _a.handlers;
                internal.hovered.delete(makeId(hoveredObj));
                if (handlers) {
                    // Clear out intersects, they are outdated by now
                    const data = Object.assign(Object.assign({}, hoveredObj), { intersections: hits || [] });
                    (_b = handlers.pointerleave) === null || _b === void 0 ? void 0 : _b.call(handlers, data);
                    (_c = handlers.pointerout) === null || _c === void 0 ? void 0 : _c.call(handlers, data);
                }
            }
        });
    }
    const handlePointer = (name) => {
        // Deal with cancelation
        switch (name) {
            case 'pointerleave':
            case 'pointercancel':
                return () => cancelPointer([]);
            case 'lostpointercapture':
                return () => {
                    eventsStateGetter().internal.captured = undefined;
                    cancelPointer([]);
                };
        }
        // Any other pointer goes here ...
        return (event) => {
            const { internal } = eventsStateGetter();
            const { objects } = instancesStateGetter();
            prepareRay(event);
            // Get fresh intersects
            const isPointerMove = name === 'pointermove';
            const filter = isPointerMove ? filterPointerEvents : undefined;
            const hits = patchIntersects(intersect(filter), event);
            // Take care of unhover
            if (isPointerMove)
                cancelPointer(hits);
            handleIntersects(hits, event, (data) => {
                var _a, _b, _c, _d;
                const eventObject = data.eventObject;
                const handlers = (_a = eventObject.__ngt) === null || _a === void 0 ? void 0 : _a.handlers;
                // Check presence of handlers
                if (!handlers)
                    return;
                if (isPointerMove) {
                    // Move event ...
                    if (handlers.pointerover ||
                        handlers.pointerenter ||
                        handlers.pointerout ||
                        handlers.pointerleave) {
                        // When enter or out is present take care of hover-state
                        const id = makeId(data);
                        const hoveredItem = internal.hovered.get(id);
                        if (!hoveredItem) {
                            // If the object wasn't previously hovered, book it and call its handler
                            internal.hovered.set(id, data);
                            (_b = handlers.pointerover) === null || _b === void 0 ? void 0 : _b.call(handlers, data);
                            (_c = handlers.pointerenter) === null || _c === void 0 ? void 0 : _c.call(handlers, data);
                        }
                        else if (hoveredItem.stopped) {
                            // If the object was previously hovered and stopped, we shouldn't allow other items to proceed
                            data.stopPropagation();
                        }
                    }
                    // Call mouse move
                    (_d = handlers.pointermove) === null || _d === void 0 ? void 0 : _d.call(handlers, data);
                }
                else {
                    // All other events ...
                    const handler = handlers === null || handlers === void 0 ? void 0 : handlers[name];
                    if (handler) {
                        // Forward all events back to their respective handlers with the exception of click events,
                        // which must use the initial target
                        if ((name !== 'click' &&
                            name !== 'contextmenu' &&
                            name !== 'dblclick') ||
                            internal.initialHits.includes(eventObject)) {
                            handler(data);
                            pointerMissed(event, Object.values(objects).filter((object) => object !== eventObject));
                        }
                    }
                }
            });
            // Save initial coordinates on pointer-down
            if (name === 'pointerdown') {
                internal.initialClick = [event.offsetX, event.offsetY];
                internal.initialHits = hits.map((hit) => hit.eventObject);
            }
            // If a click yields no results, pass it back to the user as a miss
            if ((name === 'click' || name === 'contextmenu' || name === 'dblclick') &&
                !hits.length) {
                if (calculateDistance(event) <= 2) {
                    pointerMissed(event, Object.values(objects));
                    // if (onPointerMissed) onPointerMissed();
                }
            }
        };
    };
    function pointerMissed(event, objects) {
        objects.forEach((object) => { var _a, _b, _c; return (_c = (_b = (_a = object.__ngt) === null || _a === void 0 ? void 0 : _a.handlers) === null || _b === void 0 ? void 0 : _b.pointermissed) === null || _c === void 0 ? void 0 : _c.call(_b, event); });
    }
    return { handlePointer };
}

function applyDottedPathProps(instance, key, propAtKey, isRoot = true) {
    const [first, ...paths] = key.split('.');
    const rootChild = instance[first];
    if (rootChild == null) {
        return;
    }
    if (paths.length >= 1) {
        applyDottedPathProps(rootChild, paths.join('.'), propAtKey, false);
    }
    else {
        instance[first] = propAtKey;
    }
    if (isRoot) {
        checkNeedsUpdate(rootChild);
    }
}
function checkNeedsUpdate(value) {
    if (typeof value === 'object' && 'needsUpdate' in value) {
        value['needsUpdate'] = true;
    }
}
/**
 * Apply props on instances
 *
 * @internal
 * @private
 */
function applyProps(instance, props) {
    if (!props || (props && !Object.keys(props).length))
        return;
    if ('__ngt' in props) {
        instance['__ngt'] = props['__ngt'];
    }
    if (instance['set'] != null &&
        typeof instance['set'] === 'function') {
        instance['set'](props);
    }
    for (const [key, propAtKey] of Object.entries(props)) {
        if (key.split('.').length > 1) {
            applyDottedPathProps(instance, key, propAtKey);
        }
        else {
            if (instance[key] == null) {
                instance[key] = propAtKey;
            }
            else {
                if (instance[key]['set'] !=
                    null &&
                    typeof instance[key]['set'] === 'function') {
                    if (Array.isArray(propAtKey)) {
                        instance[key]['set'](...propAtKey);
                    }
                    else {
                        instance[key]['set'](propAtKey);
                    }
                }
                else {
                    instance[key] = propAtKey;
                }
            }
            checkNeedsUpdate(propAtKey);
        }
    }
}

class ImperativeComponentStore extends ComponentStore {
    getImperativeState() {
        return this.get();
    }
}

class AnimationStore extends ImperativeComponentStore {
    constructor() {
        super({
            animations: {},
            objectSubscriptions: [],
            animationCallbacks: [],
            hasPriority: false,
        });
        this.animationsChangedEffect = this.effect((animations$) => animations$.pipe(tap((animations) => {
            const animationCallbacks = Object.values(animations);
            const hasPriority = animationCallbacks.some(({ priority }) => !!priority);
            this.patchState({ animationCallbacks, hasPriority });
        })));
        this.unregisterAnimationEffect = this.effect((uuid$) => uuid$.pipe(withLatestFrom(this.state$), tap(([uuid, { animations, objectSubscriptions }]) => {
            const subscription = objectSubscriptions.find(([objectUuid]) => objectUuid === uuid);
            if (subscription) {
                subscription[1].unsubscribe();
            }
            const _a = animations, _b = uuid, _ = _a[_b], updatedAnimations = __rest(_a, [typeof _b === "symbol" ? _b : _b + ""]);
            this.patchState({
                animations: updatedAnimations,
                objectSubscriptions: objectSubscriptions.filter(([objectUuid]) => objectUuid !== uuid),
            });
        })));
        this.clearEffect = this.effect((trigger$) => trigger$.pipe(tap(() => {
            this.patchState((state) => {
                state.objectSubscriptions.forEach(([, subscription]) => subscription.unsubscribe());
                return { animations: {}, objectSubscriptions: [] };
            });
        })));
        this.animationsChangedEffect(this.select((s) => s.animations).pipe(skip(1)));
    }
    registerAnimation(objOrObsOrCallback, callbackOrPriority, priority = 0) {
        if (objOrObsOrCallback === undefined)
            return;
        if (typeof objOrObsOrCallback === 'function') {
            const id = makeId();
            this.patchState((state) => (Object.assign(Object.assign({}, state), { animations: Object.assign(Object.assign({}, state.animations), { [id]: {
                        obj: null,
                        callback: objOrObsOrCallback,
                        priority: callbackOrPriority || 0,
                    } }) })));
            return () => {
                this.unregisterAnimationEffect(id);
            };
        }
        if (isObservable(objOrObsOrCallback)) {
            let id = '';
            const subscription = objOrObsOrCallback
                .pipe(filter((obj) => !!obj))
                .subscribe((obj) => {
                this.patchState((state) => {
                    var _a;
                    const objectSubscriptions = state.objectSubscriptions;
                    if ((_a = objectSubscriptions[objectSubscriptions.length - 1]) === null || _a === void 0 ? void 0 : _a[0]) {
                        objectSubscriptions[objectSubscriptions.length - 1][0] = obj.uuid;
                    }
                    id = obj.uuid;
                    return {
                        animations: Object.assign(Object.assign({}, state.animations), { [id]: {
                                obj,
                                callback: callbackOrPriority,
                                priority,
                            } }),
                        objectSubscriptions,
                    };
                });
            });
            this.patchState((state) => ({
                objectSubscriptions: [
                    ...state.objectSubscriptions,
                    [null, subscription],
                ],
            }));
            return () => {
                this.unregisterAnimationEffect(id);
            };
        }
        const uuid = objOrObsOrCallback.uuid;
        this.patchState((state) => ({
            animations: Object.assign(Object.assign({}, state.animations), { [uuid]: {
                    obj: objOrObsOrCallback,
                    callback: callbackOrPriority,
                    priority,
                } }),
        }));
        return () => {
            this.unregisterAnimationEffect(uuid);
        };
    }
    ngOnDestroy() {
        this.clearEffect();
        super.ngOnDestroy();
    }
}
AnimationStore.decorators = [
    { type: Injectable }
];
AnimationStore.ctorParameters = () => [];

class CanvasStore extends ImperativeComponentStore {
    constructor() {
        super({
            isOrthographic: false,
            isLinear: false,
            shadows: false,
            mouse: new Vector2(),
            clock: new Clock(),
            internal: {
                active: false,
                size: { width: 0, height: 0 },
                dpr: 1,
            },
        });
        this.renderer$ = this.select((s) => s.renderer);
        this.camera$ = this.select((s) => s.camera);
        this.scene$ = this.select((s) => s.scene);
        this.raycaster$ = this.select((s) => s.raycaster);
        this.isOrthographic$ = this.select((s) => s.isOrthographic);
        this.isLinear$ = this.select((s) => s.isLinear);
        this.shadows$ = this.select((s) => s.shadows);
        this.active$ = this.select((s) => s.internal.active);
        this.canvasInternal$ = this.select((s) => s.internal);
        this.setIsOrthographic = this.updater((state, isOrthographic) => (Object.assign(Object.assign({}, state), { isOrthographic })));
        this.setIsLinear = this.updater((state, isLinear) => (Object.assign(Object.assign({}, state), { isLinear })));
        this.setShadows = this.updater((state, shadows) => (Object.assign(Object.assign({}, state), { shadows })));
        this.setSize = this.updater((state, size) => (Object.assign(Object.assign({}, state), { internal: Object.assign(Object.assign({}, state.internal), { size }) })));
        this.setDpr = this.updater((state, dpr) => (Object.assign(Object.assign({}, state), { internal: Object.assign(Object.assign({}, state.internal), { dpr }) })));
        this.setActive = this.updater((state, active) => (Object.assign(Object.assign({}, state), { internal: Object.assign(Object.assign({}, state.internal), { active }) })));
        this.initRendererEffect = this.effect((canvas$) => {
            return canvas$.pipe(withLatestFrom(this.canvasInternal$, this.isLinear$, this.shadows$), tap(([canvas, { size, dpr }, isLinear, shadows]) => {
                const renderer = new WebGLRenderer({
                    canvas,
                    antialias: true,
                    powerPreference: 'high-performance',
                    alpha: true,
                });
                if (shadows) {
                    renderer.shadowMap.enabled = true;
                    if (typeof shadows === 'object')
                        Object.assign(renderer.shadowMap, shadows);
                    else
                        renderer.shadowMap.type = PCFSoftShadowMap;
                }
                if (!isLinear) {
                    renderer.toneMapping = ACESFilmicToneMapping;
                    renderer.outputEncoding = sRGBEncoding;
                }
                renderer.setClearAlpha(0);
                renderer.setPixelRatio(dpr);
                renderer.setSize(size.width, size.height);
                this.patchState({ renderer });
            }));
        });
        this.initSceneEffect = this.effect((sceneOptions$) => sceneOptions$.pipe(tap((sceneOptions) => {
            const scene = new Scene();
            applyProps(scene, sceneOptions);
            this.patchState({ scene });
        })));
        this.initCameraEffect = this.effect((cameraOptions$) => cameraOptions$.pipe(withLatestFrom(this.isOrthographic$, this.canvasInternal$), tap(([cameraOptions, isOrthographic, { size }]) => {
            var _a;
            const isCamera = cameraOptions instanceof Camera;
            let camera;
            if (isCamera) {
                camera = cameraOptions;
            }
            else {
                if (isOrthographic) {
                    camera = new OrthographicCamera(0, 0, 0, 0, 0.1, 1000);
                    camera.zoom = 100;
                }
                else {
                    camera = new PerspectiveCamera(75, (_a = size.width / size.height) !== null && _a !== void 0 ? _a : 0, 0.1, 1000);
                }
                camera.position.z = 5;
                applyProps(camera, cameraOptions);
            }
            // look at center by default
            camera.lookAt(0, 0, 0);
            this.patchState({ camera });
        })));
        this.initRaycasterEffect = this.effect((raycasterOptions$) => raycasterOptions$.pipe(tap((raycasterOptions) => {
            const raycaster = new Raycaster();
            raycaster.enabled = true;
            applyProps(raycaster, raycasterOptions);
            this.patchState({ raycaster });
        })));
        this.windowResizeEffect = this.effect((params$) => params$.pipe(withLatestFrom(this.renderer$, this.camera$), tap(([{ dpr, size }, renderer, camera]) => {
            if (camera.type === 'PerspectiveCamera') {
                camera.aspect = size.width / size.height;
            }
            else {
                camera.left = size.width / -2;
                camera.right = size.width / 2;
                camera.top = size.height / 2;
                camera.bottom = size.height / -2;
            }
            camera.updateProjectionMatrix();
            camera.updateMatrixWorld();
            renderer.setPixelRatio(dpr);
            renderer.setSize(size.width, size.height);
            this.setSize(size);
            this.setDpr(dpr);
        })));
        this.setActive(combineLatest([
            this.renderer$,
            this.camera$,
            this.scene$,
            this.raycaster$,
        ]).pipe(filter((coreObjects) => coreObjects.every(Boolean)), mapTo(true)));
    }
    ngOnDestroy() {
        const { renderer } = this.getImperativeState();
        if (renderer) {
            renderer.renderLists.dispose();
            renderer.forceContextLoss();
        }
        super.ngOnDestroy();
    }
}
CanvasStore.decorators = [
    { type: Injectable }
];
CanvasStore.ctorParameters = () => [];

class InstancesStore extends ImperativeComponentStore {
    constructor() {
        super({ materials: {}, bufferGeometries: {}, objects: {} });
        this.saveMaterial = this.updater((state, { material, id = material.uuid }) => (Object.assign(Object.assign({}, state), { materials: Object.assign(Object.assign({}, state.materials), { [id]: material }) })));
        this.removeMaterial = this.updater((state, id) => {
            const _a = state.materials, _b = id, _ = _a[_b], materials = __rest(_a, [typeof _b === "symbol" ? _b : _b + ""]);
            return Object.assign(Object.assign({}, state), { materials });
        });
        this.saveBufferGeometry = this.updater((state, { bufferGeometry, id = bufferGeometry.uuid }) => (Object.assign(Object.assign({}, state), { bufferGeometries: Object.assign(Object.assign({}, state.bufferGeometries), { [id]: bufferGeometry }) })));
        this.removeBufferGeometry = this.updater((state, id) => {
            const _a = state.bufferGeometries, _b = id, _ = _a[_b], bufferGeometries = __rest(_a, [typeof _b === "symbol" ? _b : _b + ""]);
            return Object.assign(Object.assign({}, state), { bufferGeometries });
        });
        this.saveObject = this.updater((state, obj) => (Object.assign(Object.assign({}, state), { objects: Object.assign(Object.assign({}, state.objects), { [obj.uuid]: obj }) })));
        this.removeObject = this.updater((state, id) => {
            const _a = state.objects, _b = id, _ = _a[_b], objects = __rest(_a, [typeof _b === "symbol" ? _b : _b + ""]);
            return Object.assign(Object.assign({}, state), { objects });
        });
    }
}
InstancesStore.decorators = [
    { type: Injectable }
];
InstancesStore.ctorParameters = () => [];

const supportedEvents = [
    'click',
    'contextmenu',
    'dblclick',
    'wheel',
    'pointerdown',
    'pointerup',
    'pointerleave',
    'pointermove',
    'pointercancel',
    'lostpointercapture',
];
class EventsStore extends ImperativeComponentStore {
    constructor(canvasStore, instancesStore) {
        super({
            connected: false,
            internal: {
                interaction: [],
                hovered: new Map(),
                captured: undefined,
                initialClick: [0, 0],
                initialHits: [],
            },
        });
        this.canvasStore = canvasStore;
        this.instancesStore = instancesStore;
        this.initHandlersEffect = this.effect((trigger$) => trigger$.pipe(tap(() => {
            const { handlePointer } = createEvents(() => this.canvasStore.getImperativeState(), () => this.getImperativeState(), () => this.instancesStore.getImperativeState());
            this.patchState({
                handlers: supportedEvents.reduce((handlers, supportedEventName) => (Object.assign(Object.assign({}, handlers), { [supportedEventName]: handlePointer(supportedEventName) })), {}),
            });
        })));
        this.connectEffect = this.effect((target$) => target$.pipe(tap((target) => {
            this.disconnectEffect();
            const { handlers } = this.getImperativeState();
            Object.entries(handlers !== null && handlers !== void 0 ? handlers : []).forEach(([name, event]) => target.addEventListener(name, event, { passive: true }));
            this.patchState({ connected: target });
        })));
        this.disconnectEffect = this.effect((trigger$) => trigger$.pipe(tap(() => {
            const { handlers, connected } = this.getImperativeState();
            if (connected) {
                Object.entries(handlers !== null && handlers !== void 0 ? handlers : []).forEach(([name, event]) => {
                    if (connected instanceof HTMLElement) {
                        connected.removeEventListener(name, event);
                    }
                });
                this.patchState({ connected: false });
            }
        })));
        this.initHandlersEffect();
    }
    ngOnDestroy() {
        this.disconnectEffect();
        super.ngOnDestroy();
    }
}
EventsStore.decorators = [
    { type: Injectable }
];
EventsStore.ctorParameters = () => [
    { type: CanvasStore },
    { type: InstancesStore }
];

class LoopService {
    constructor(canvasStore, animationStore) {
        this.canvasStore = canvasStore;
        this.animationStore = animationStore;
        this.loopClock = new Clock();
    }
    start() {
        const { renderer } = this.canvasStore.getImperativeState();
        if (renderer) {
            renderer.setAnimationLoop(() => {
                this.tick(this.loopClock.getDelta());
            });
        }
    }
    stop() {
        const { renderer } = this.canvasStore.getImperativeState();
        if (renderer) {
            renderer.setAnimationLoop(null);
        }
    }
    tick(delta) {
        const { renderer, scene, camera, internal, clock, } = this.canvasStore.getImperativeState();
        const { animationCallbacks, hasPriority, } = this.animationStore.getImperativeState();
        if (renderer && scene && camera) {
            if (hasPriority) {
                animationCallbacks.sort(({ priority: a = 0 }, { priority: b = 0 }) => a - b);
            }
            else {
                renderer.render(scene, camera);
            }
            for (const animationCallback of animationCallbacks) {
                if (animationCallback.obj) {
                    animationCallback.callback(animationCallback.obj, {
                        clock,
                        camera,
                        scene,
                        renderer,
                        size: internal.size,
                        delta,
                    });
                }
                else {
                    animationCallback.callback({
                        clock,
                        camera,
                        scene,
                        renderer,
                        size: internal.size,
                        delta,
                    });
                }
            }
        }
    }
    ngOnDestroy() {
        this.stop();
    }
}
LoopService.decorators = [
    { type: Injectable }
];
LoopService.ctorParameters = () => [
    { type: CanvasStore },
    { type: AnimationStore }
];

class DestroyedService extends Subject {
    ngOnDestroy() {
        this.next();
        this.complete();
    }
}
DestroyedService.decorators = [
    { type: Injectable }
];

class AnimationLoopParticipant {
    constructor(animationStore, ngZone) {
        this.animationStore = animationStore;
        this.ngZone = ngZone;
        this.animateReady = new EventEmitter();
    }
    participate(animateObject) {
        this.ngZone.runOutsideAngular(() => {
            if (this.animateReady.observers.length) {
                if (animateObject instanceof Object3D) {
                    this.animateTeardown = this.animationStore.registerAnimation(animateObject, (obj, state) => {
                        this.animateReady.emit({
                            animateObject: obj,
                            renderState: state,
                        });
                    });
                }
                else {
                    this.animateTeardown = this.animationStore.registerAnimation((state) => {
                        this.animateReady.emit({
                            animateObject,
                            renderState: state,
                        });
                    });
                }
            }
        });
    }
    ngOnDestroy() {
        this.ngZone.runOutsideAngular(() => {
            if (this.animateTeardown) {
                this.animateTeardown();
            }
        });
    }
}
AnimationLoopParticipant.decorators = [
    { type: Directive }
];
AnimationLoopParticipant.ctorParameters = () => [
    { type: AnimationStore },
    { type: NgZone }
];
AnimationLoopParticipant.propDecorators = {
    animateReady: [{ type: Output }]
};

class ThreeObject3d extends AnimationLoopParticipant {
    constructor(canvasStore, instancesStore, eventsStore, animationStore, ngZone, parentObjectDirective) {
        super(animationStore, ngZone);
        this.canvasStore = canvasStore;
        this.instancesStore = instancesStore;
        this.eventsStore = eventsStore;
        this.animationStore = animationStore;
        this.ngZone = ngZone;
        this.parentObjectDirective = parentObjectDirective;
        this.castShadow = false;
        this.receiveShadow = false;
        this.visible = true;
        this.matrixAutoUpdate = true;
        this.appendMode = 'immediate';
        // events
        this.click = new EventEmitter();
        this.contextmenu = new EventEmitter();
        this.dblclick = new EventEmitter();
        this.pointerup = new EventEmitter();
        this.pointerdown = new EventEmitter();
        this.pointerover = new EventEmitter();
        this.pointerout = new EventEmitter();
        this.pointerenter = new EventEmitter();
        this.pointerleave = new EventEmitter();
        this.pointermove = new EventEmitter();
        this.pointermissed = new EventEmitter();
        this.pointercancel = new EventEmitter();
        this.wheel = new EventEmitter();
        this.$object3d = new BehaviorSubject(null);
        this.ready = this.object3d$.pipe(filter(Boolean));
    }
    get object3d$() {
        return this.$object3d.asObservable();
    }
    ngOnChanges(changes) {
        if (this.object3d) {
            this.applyCustomProps(changes);
        }
    }
    init() {
        this.ngZone.runOutsideAngular(() => {
            this.initObject();
            if (this.object3d) {
                this.applyCustomProps();
                applyProps(this.object3d, {
                    __ngt: {
                        canvasStateGetter: () => this.canvasStore.getImperativeState(),
                        eventsStateGetter: () => this.eventsStore.getImperativeState(),
                        handlers: this.applyEvents(),
                    },
                });
                this.instancesStore.saveObject(this.object3d);
                this.appendToParent();
                this.objectReady();
            }
        });
    }
    addToScene() {
        const { scene } = this.canvasStore.getImperativeState();
        if (scene) {
            scene.add(this.object3d);
        }
    }
    addToParent() {
        if (this.parentObjectDirective) {
            this.parentObjectDirective.object3d.add(this.object3d);
        }
        else {
            this.addToScene();
        }
    }
    remove() {
        if (this.appendTo) {
            this.appendTo.remove(this.object3d);
        }
        else if (this.parentObjectDirective && this.appendMode === 'immediate') {
            this.parentObjectDirective.object3d.remove(this.object3d);
        }
        else {
            const { scene } = this.canvasStore.getImperativeState();
            if (scene) {
                scene.remove(this.object3d);
            }
        }
        this.object3d.clear();
    }
    objectReady() {
        this.$object3d.next(this.object3d);
        this.participate(this.object3d);
    }
    applyCustomProps(inputChanges) {
        this.ngZone.runOutsideAngular(() => {
            const customProps = {
                castShadow: this.castShadow,
                receiveShadow: this.receiveShadow,
                visible: this.visible,
                matrixAutoUpdate: this.matrixAutoUpdate,
            };
            if (this.name) {
                customProps['name'] = this.name;
            }
            if (this.position) {
                customProps['position'] = this.position;
            }
            if (this.rotation) {
                customProps['rotation'] = this.rotation;
            }
            else if (this.quaternion) {
                customProps['quaternion'] = this.quaternion;
            }
            if (this.scale) {
                customProps['scale'] = this.scale;
            }
            if (this.userData) {
                customProps['userData'] = this.userData;
            }
            if (this.color) {
                this.color = Array.isArray(this.color)
                    ? new Color(...this.color)
                    : new Color(this.color);
                if (!this.canvasStore.getImperativeState().isLinear) {
                    this.color.convertSRGBToLinear();
                }
                customProps['color'] = this.color;
            }
            if (inputChanges) {
                for (const [inputName, inputChange] of Object.entries(inputChanges)) {
                    if (!inputChange.isFirstChange() ||
                        [
                            'name',
                            'position',
                            'rotation',
                            'quaternion',
                            'scale',
                            'userData',
                            'color',
                        ].includes(inputName) // skip 7 common inputs
                    ) {
                        continue;
                    }
                    customProps[inputName] = inputChange.currentValue;
                }
            }
            applyProps(this.object3d, customProps);
            this.object3d.updateMatrix();
        });
    }
    applyEvents() {
        const handlers = {};
        [
            'click',
            'contextmenu',
            'dblclick',
            'pointerup',
            'pointerdown',
            'pointerover',
            'pointerout',
            'pointerenter',
            'pointerleave',
            'pointermove',
            'pointermissed',
            'pointercancel',
            'wheel',
        ].forEach((eventName) => {
            if (this[eventName].observers.length) {
                handlers[eventName] = (event) => {
                    this.ngZone.run(() => {
                        this[eventName].emit(event);
                    });
                };
            }
        });
        return handlers;
    }
    appendToParent() {
        if (this.appendTo) {
            this.appendTo.add(this.object3d);
            return;
        }
        if (this.appendMode === 'root') {
            this.addToScene();
            return;
        }
        if (this.appendMode === 'immediate') {
            this.addToParent();
        }
    }
    ngOnDestroy() {
        super.ngOnDestroy();
        this.ngZone.runOutsideAngular(() => {
            if (this.object3d) {
                this.remove();
                this.instancesStore.removeObject(this.object3d.uuid);
                this.animationStore.unregisterAnimationEffect(this.object3d.uuid);
            }
        });
    }
}
ThreeObject3d.decorators = [
    { type: Directive }
];
ThreeObject3d.ctorParameters = () => [
    { type: CanvasStore },
    { type: InstancesStore },
    { type: EventsStore },
    { type: AnimationStore },
    { type: NgZone },
    { type: ThreeObject3d, decorators: [{ type: Optional }, { type: SkipSelf }] }
];
ThreeObject3d.propDecorators = {
    name: [{ type: Input }],
    position: [{ type: Input }],
    rotation: [{ type: Input }],
    quaternion: [{ type: Input }],
    scale: [{ type: Input }],
    color: [{ type: Input }],
    userData: [{ type: Input }],
    castShadow: [{ type: Input }],
    receiveShadow: [{ type: Input }],
    visible: [{ type: Input }],
    matrixAutoUpdate: [{ type: Input }],
    appendMode: [{ type: Input }],
    appendTo: [{ type: Input }],
    click: [{ type: Output }],
    contextmenu: [{ type: Output }],
    dblclick: [{ type: Output }],
    pointerup: [{ type: Output }],
    pointerdown: [{ type: Output }],
    pointerover: [{ type: Output }],
    pointerout: [{ type: Output }],
    pointerenter: [{ type: Output }],
    pointerleave: [{ type: Output }],
    pointermove: [{ type: Output }],
    pointermissed: [{ type: Output }],
    pointercancel: [{ type: Output }],
    wheel: [{ type: Output }],
    ready: [{ type: Output }]
};

class ThreeObject3dMaterialGeometry extends ThreeObject3d {
    set extraArgs(v) {
        this._extraArgs = v;
    }
    ngAfterContentInit() {
        this.init();
    }
    initObject() {
        if (this.canCreate()) {
            const material = this.getMaterial();
            const geometry = this.getGeometry();
            this._objectMaterialGeometry = new this.objectType(geometry, material, ...(this._extraArgs || []));
            if (this.morphTargetDictionary &&
                'morphTargetDictionary' in this._objectMaterialGeometry) {
                this
                    ._objectMaterialGeometry.morphTargetDictionary = this.morphTargetDictionary;
            }
            if (this.morphTargetInfluences &&
                'morphTargetInfluences' in this._objectMaterialGeometry) {
                this
                    ._objectMaterialGeometry.morphTargetInfluences = this.morphTargetInfluences;
            }
            if (this.customize) {
                this.customize();
            }
        }
    }
    canCreate() {
        return true;
    }
    get object3d() {
        return this._objectMaterialGeometry;
    }
    getMaterial() {
        if (this.material) {
            if ((Array.isArray(this.material) &&
                this.material[0] instanceof Material) ||
                this.material instanceof Material) {
                return this.material;
            }
            if (Array.isArray(this.material)) {
                return this.material.map((materialId) => this.instancesStore.getImperativeState().materials[materialId]);
            }
            return this.instancesStore.getImperativeState().materials[this.material];
        }
        return undefined;
    }
    getGeometry() {
        if (this.geometry) {
            if (this.geometry instanceof BufferGeometry) {
                return this.geometry;
            }
            return this.instancesStore.getImperativeState().bufferGeometries[this.geometry];
        }
        return undefined;
    }
}
ThreeObject3dMaterialGeometry.decorators = [
    { type: Directive }
];
ThreeObject3dMaterialGeometry.propDecorators = {
    geometry: [{ type: Input }],
    material: [{ type: Input }],
    morphTargetInfluences: [{ type: Input }],
    morphTargetDictionary: [{ type: Input }]
};

/**
 * Generated bundle index. Do not edit.
 */

export { AnimationLoopParticipant, AnimationStore, CanvasStore, DestroyedService, EventsStore, InstancesStore, LoopService, ThreeObject3d, ThreeObject3dMaterialGeometry, applyProps };
//# sourceMappingURL=angular-three-core.js.map
